import os
import os.path as opath
import logging

from .exceptions import UnacknowledgedOption, BadRequest


"""
This module adds methods that parse TFTP read/write requests, TFTP options,
as well as parse filenames.
"""


def validate_req(fname, mode, opts, supported_opts=None, default_opts=None):
    """
    Validates an RRQ or WRQ.
    Currently only validates the optional "opts" parameter.
    """
    if not supported_opts:
        supported_opts = {}
    if not default_opts:
        default_opts = {}

    acknowledged_options = {}

    for option, value in opts.items():
        logging.debug(option)
        if option in supported_opts.keys():
            logging.debug(option)
            try:
                acknowledged_options[option] = supported_opts[option](value)
            except UnacknowledgedOption as e:
                logging.debug(e)
            except ValueError:
                logging.debug(
                    ('Client passed malformed option "{0}": "{1}", '
                     'ignoring').format(option, value))

    return (fname.decode(encoding='ascii'), mode, acknowledged_options)


def parse_req(req):
    """
    Seperates \x00 delimited byte string contents according to RFC2347:
    'filename\x00mode\x00opt1\x00val1\x00optN\x00valN\x00' into a
    filename, a mode, and a dictionary of option:values.
    """
    logging.debug("Request: {}".format(req))
    try:
        fname, mode, *opts = [item for item in req.split(b'\x00') if item]
    except ValueError:
        raise BadRequest("Could not parse request: {}".format(req))
    options = dict(zip(opts[::2], opts[1::2]))
    return fname, mode, options


def blksize_parser(val, lower_bound=8, upper_bound=65464):
    """
    Parses and validates the 'blksize' option against the RFC 2348.
    """
    value = int(val)
    if value > upper_bound:
        return value - (value - upper_bound)
    elif value < lower_bound:
        raise UnacknowledgedOption(
            'Requested blksize "{0}" below RFC-spec limit ({1})'.format(
                value, lower_bound))
    else:
        return value


def timeout_parser(val, lower_bound=1, upper_bound=255):
    """
    Parses and validates the 'timeout' option against RFC 2349.
    """
    value = float(val)

    if value > upper_bound or value < lower_bound:
        raise UnacknowledgedOption(
            ('Requested timeout "{0}" outside of RFC-spec range '
             '({1} - {2})').format(value, lower_bound, upper_bound))
    else:
        return value



import logging
import asyncio

from .protocols import TFTPServerProtocol
from .cli_parser import parse_cli_arguments


def main():
    args = parse_cli_arguments()

    logging.info('Starting TFTP server on {addr}:{port}'.format(
        addr=args.host, port=args.port))

    timeouts = {k: v for k, v in vars(args).items() if 'timeout' in k}
    loop = asyncio.get_event_loop()

    listen = loop.create_datagram_endpoint(
        lambda: TFTPServerProtocol(args.host, loop, timeouts),
        local_addr=(args.host, args.port,))

    transport, protocol = loop.run_until_complete(listen)

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        logging.info('Received signal, shutting down')

    transport.close()
    loop.close()

if __name__ == '__main__':
    main()



import os
import os.path as opath


def sanitize_fname(fname):
    """
    Ensures that fname is a path under the current working directory.
    """
    root_dir = os.getcwd()
    return opath.join(
        bytes(root_dir, encoding='ascii'),
        opath.normpath(
            b'/' + fname).lstrip(b'/'))


class FileReader(object):
    """
    A wrapper around a regular file that implements:
    - read_chunk - for closing the file when bytes read is
      less than chunk_size.
    - finished - for easier notifications
    interfaces.
    When it goes out of scope, it ensures the file is closed.
    """

    def __init__(self, fname, chunk_size=0):
        self.fname = sanitize_fname(fname)
        self.chunk_size = chunk_size
        self._f = None
        self._f = self._open_file()
        self.finished = False

    def _open_file(self):
        return open(self.fname, 'rb')

    def read_chunk(self, size=None):
        size = size or self.chunk_size
        if self.finished:
            return b''

        data = self._f.read(size)

        if not data or (size > 0 and len(data) < size):
            self._f.close()
            self.finished = True

        return data

    def __del__(self):
        if self._f and not self._f.closed:
            self._f.close()


class FileWriter(object):
    """
    Wrapper around a regular file that implements:
    - write_chunk - for closing the file when bytes written
      is less than chunk_size.
    When it goes out of scope, it ensures the file is closed.
    """
    def __init__(self, fname, chunk_size):
        self.fname = fname
        self.chunk_size = chunk_size
        self._f = None
        self._f = self._open_file()

    def _open_file(self):
        return open(self.fname, 'xb')

    def _flush(self):
        if self._f:
            self._f.flush()

    def write_chunk(self, data):
        bytes_written = self._f.write(data)

        if not data or len(data) < self.chunk_size:
            self._f.close()

        return bytes_written

    def __del__(self):
        if self._f and not self._f.closed:
            self._f.close()



import argparse
import logging
from sys import exit

from py3tftp import __version__

EPILOG = """
Released under the MIT license.
Copyright 2016 Matt O. <matt@mattscodecave.com>
"""

logging_config = {
    'format': '%(asctime)s [%(levelname)s] %(message)s',
    'level': logging.INFO,
    'filename': None
}


def print_version():
    print("py3tftp version: {}".format(__version__))


def parse_cli_arguments():
    parser = argparse.ArgumentParser(epilog=EPILOG)
    parser.add_argument('--host',
                        default='0.0.0.0',
                        help=('IP of the interface the server will listen on. '
                              'Default: 0.0.0.0'))
    parser.add_argument(
        '-p',
        '--port',
        default=9069,
        type=int,
        help=('Port the server will listen on. '
              'Default: 9069. TFTP standard-compliant port: 69 - '
              'requires superuser privileges.'))
    parser.add_argument(
        '--ack-timeout',
        default=0.5,
        type=float,
        help='Timeout for each ACK of the lock-step. Default: 0.5.')
    parser.add_argument(
        '--conn-timeout',
        dest="timeout",
        default=3.0,
        type=float,
        help=('Timeout before the server gives up on a transfer and closes '
              'the connection. Default: 3.'))
    parser.add_argument('-l', '--file-log', help='Append output to log file.')
    parser.add_argument('-v',
                        '--verbose',
                        action='store_true',
                        help='Enable debug-level logging.')
    parser.add_argument('--version', action='store_true')

    args = parser.parse_args()

    if args.verbose:
        logging_config['level'] = logging.DEBUG

    if args.file_log:
        logging_config['filename'] = args.log

    if args.version:
        print_version()
        exit()

    logging.basicConfig(**logging_config)
    return args



from py3tftp import tftp_parsing
from py3tftp.exceptions import BadRequest, BadPacketType


class TFTPPacketFactory(object):
    def __init__(self, supported_opts=None, default_opts=None):
        self.supported_opts = supported_opts or {}
        self.default_opts = default_opts or {}

    @classmethod
    def create_packet(cls, pkt_type=None, **kwargs):
        if pkt_type in ('RRQ', 'WRQ'):
            return TFTPRequestPacket(pkt_type, **kwargs)
        elif pkt_type == 'DAT':
            return TFTPDatPacket(**kwargs)
        elif pkt_type == 'ACK':
            return TFTPAckPacket(**kwargs)
        elif pkt_type == 'OCK':
            return TFTPOckPacket(**kwargs)
        elif pkt_type == 'ERR':
            return TFTPErrPacket(**kwargs)
        raise BadPacketType('Unknown packet type - "{0}"'.format(pkt_type))

    def from_bytes(self, data):
        try:
            pkt_type = BaseTFTPPacket.pkt_types[data[:2]].upper()
        except KeyError:
            raise BadPacketType(
                'Cannot create packet from raw bytes, unknown packet type.')

        if pkt_type in ('RRQ', 'WRQ'):
            fname, mode, r_opts = tftp_parsing.validate_req(
                *tftp_parsing.parse_req(data[2:]),
                supported_opts=self.supported_opts,
                default_opts=self.default_opts)
            return self.create_packet(
                pkt_type=pkt_type,
                fname=fname,
                mode=mode,
                r_opts=r_opts)
        elif pkt_type == 'DAT':
            block_no = BaseTFTPPacket.unpack_short(data[2:4])
            return self.create_packet(
                pkt_type=pkt_type,
                block_no=block_no,
                data=data[4:])
        elif pkt_type == 'ACK':
            block_no = BaseTFTPPacket.unpack_short(data[2:4])
            return self.create_packet(
                pkt_type=pkt_type,
                block_no=block_no)
        elif pkt_type == 'OCK':
            _, _, r_opts = tftp_parsing.validate_req(
                *tftp_parsing.parse_req(data[2:]),
                supported_opts=self.supported_opts,
                default_opts=self.default_opts)
            return self.create_packet(pkt_type=pkt_type, opts=r_opts)
        elif pkt_type == 'ERR':
            code = BaseTFTPPacket.unpack_short(data[2:4])
            msg = data[4:]
            return self.create_packet(pkt_type, code=code, msg=msg)

    @classmethod
    def err_file_exists(cls):
        return cls.create_packet('ERR', code=6, msg='File already exists')

    @classmethod
    def err_access_violation(cls):
        return cls.create_packet('ERR', code=2, msg='Access violation')

    @classmethod
    def err_file_not_found(cls):
        return cls.create_packet('ERR', code=1, msg='File not found')

    @classmethod
    def err_unknown_tid(cls):
        return cls.create_packet('ERR', code=5, msg='Unknown transfer id')


class BaseTFTPPacket(object):
    pkt_types = {
        b'\x00\x01': 'RRQ',
        b'\x00\x02': 'WRQ',
        b'\x00\x03': 'DAT',
        b'\x00\x04': 'ACK',
        b'\x00\x05': 'ERR',
        b'\x00\x06': 'OCK',
        'RRQ': b'\x00\x01',
        'WRQ': b'\x00\x02',
        'DAT': b'\x00\x03',
        'ACK': b'\x00\x04',
        'ERR': b'\x00\x05',
        'OCK': b'\x00\x06',
    }

    def __init__(self):
        self.pkt_type = None
        self._bytes_cache = None

    def to_bytes(self):
        raise NotImplementedError

    def is_ack(self):
        return self.pkt_type == 'ACK'

    def is_correct_sequence(self, expected_block_no):
        """
        Checks whether incoming data packet has the expected block number.
        """
        return expected_block_no == self.block_no

    def is_data(self):
        return self.pkt_type == 'DAT'

    def is_err(self):
        return self.pkt_type == 'ERR'

    def is_rrq(self):
        return self.pkt_type == 'RRQ'

    def is_wrq(self):
        return self.pkt_type == 'WRQ'

    def is_ock(self):
        return self.pkt_type == 'OCK'

    @property
    def size(self):
        return len(self.to_bytes())

    @classmethod
    def _to_bytes(cls, item):
        if isinstance(item, bytes):
            return item
        else:
            return str(item).encode('ascii')

    @classmethod
    def serialize_options(cls, options):
        opt_items = [val for pair in options.items() for val in pair]
        opt_items = [cls._to_bytes(val) for val in opt_items]
        return b'\x00'.join(opt_items)

    @classmethod
    def number_to_bytes(cls, val):
        """
        Changes a number to an ascii byte string.
        """
        return bytes(str(int(val)), encoding='ascii')

    @classmethod
    def pack_short(cls, number):
        """
        Create big-endian short byte string out of integer.
        """
        return number.to_bytes(2, byteorder='big')

    @classmethod
    def unpack_short(cls, data):
        """
        Create integer out of big-endian short byte string.
        """
        return int.from_bytes(data, byteorder='big')


def text_to_bytes(string):
    if not isinstance(string, bytes):
        return bytes(string, encoding='ascii')
    return string


class TFTPRequestPacket(BaseTFTPPacket):
    def __init__(self, pkt_type, **kwargs):
        super().__init__()
        self.pkt_type = pkt_type.upper()
        self.fname = text_to_bytes(kwargs['fname'])
        self.mode = text_to_bytes(kwargs['mode'])
        self.r_opts = kwargs.get('r_opts', {})

    def to_bytes(self):
        packet = [self.pkt_types[self.pkt_type] + self.fname,
                  self.mode,
                  BaseTFTPPacket.serialize_options(self.r_opts)]

        return b'\x00'.join([part for part in packet if part]) + b'\x00'


class TFTPAckPacket(BaseTFTPPacket):
    def __init__(self, **kwargs):
        super().__init__()
        self.pkt_type = 'ACK'
        self.block_no = kwargs['block_no']

    def to_bytes(self):
        return b''.join([self.pkt_types['ACK'],
                         BaseTFTPPacket.pack_short(self.block_no)])


class TFTPDatPacket(BaseTFTPPacket):
    def __init__(self, **kwargs):
        super().__init__()
        self.pkt_type = 'DAT'

        self.block_no = kwargs['block_no']
        self.data = kwargs['data']

    def to_bytes(self):
        return b''.join([self.pkt_types['DAT'],
                         BaseTFTPPacket.pack_short(self.block_no),
                         self.data])


class TFTPOckPacket(BaseTFTPPacket):
    def __init__(self, **kwargs):
        super().__init__()
        self.pkt_type = 'OCK'
        self.options = kwargs.get('r_opts', {})

    def to_bytes(self):
        return b''.join([self.pkt_types['OCK'],
                         BaseTFTPPacket.serialize_options(self.options),
                         b'\x00'])


class TFTPErrPacket(BaseTFTPPacket):
    def __init__(self, **kwargs):
        super().__init__()
        self.pkt_type = 'ERR'
        self.code = kwargs['code']
        self.msg = kwargs['msg']

    def to_bytes(self):
        return b''.join([self.pkt_types['ERR'],
                         BaseTFTPPacket.pack_short(self.code),
                         self.msg.encode('ascii'),
                         b'\x00'])



import logging
import asyncio

from .exceptions import ProtocolException
from py3tftp import tftp_parsing
from py3tftp import file_io
from py3tftp.tftp_packet import TFTPPacketFactory


class BaseTFTPProtocol(asyncio.DatagramProtocol):
    supported_opts = {
        b'blksize': tftp_parsing.blksize_parser,
        b'timeout': tftp_parsing.timeout_parser,
    }

    default_opts = {b'ack_timeout': 0.5, b'timeout': 5.0, b'blksize': 512}

    def __init__(self, packet, file_handler_cls, remote_addr, extra_opts=None):
        self.packet_factory = TFTPPacketFactory(
            supported_opts=self.supported_opts,
            default_opts=self.default_opts)

        self.remote_addr = remote_addr
        self.packet = self.packet_factory.from_bytes(packet)
        self.extra_opts = extra_opts or {}
        self.file_handler_cls = file_handler_cls
        self.retransmit = None
        self.file_handler = None
        self.finished = False

    def datagram_received(self, data, addr):
        """
        Processes every received datagram.
        """
        raise NotImplementedError

    def initialize_transfer(self):
        """
        Sets up the message counter and attempts to open the target file for
        reading or writing.
        """
        raise NotImplementedError

    def next_datagram(self):
        """
        Returns the next datagram to be sent to self.remote_addr.
        """
        raise NotImplementedError

    def connection_made(self, transport):
        """
        Triggers connection initialization at the beginning of a connection.
        """
        self.transport = transport
        self.handle_initialization()

    def handle_initialization(self):
        """
        Sends first packet to self.remote_addr. In the process, it attempts to
        access the requested file - and handles possible file errors - as well
        as handling option negotiation (if applicable).
        """
        try:
            self.set_proto_attributes()
            self.initialize_transfer()

            if self.r_opts:
                self.counter = 0
                pkt = self.packet_factory.create_packet('OCK',
                                                        r_opts=self.r_opts)
            else:
                pkt = self.next_datagram()
        except FileExistsError:
            logging.error('"{}" already exists! Cannot overwrite'.format(
                self.filename))
            pkt = self.packet_factory.err_file_exists()
        except PermissionError:
            logging.error('Insufficient permissions to operate on "{}"'.format(
                self.filename))
            pkt = self.packet_factory.err_access_violation()
        except FileNotFoundError:
            logging.error('File "{}" does not exist!'.format(self.filename))
            pkt = self.packet_factory.err_file_not_found()

        logging.debug('opening pkt: {}'.format(pkt))
        self.send_opening_packet(pkt.to_bytes())

        if pkt.is_err():
            self.handle_err_pkt()

    def set_proto_attributes(self):
        """
        Sets the self.filename , self.opts, and self.r_opts.
        The caller should handle any exceptions and react accordingly
        ie. send error packet, close connection, etc.
        """
        self.filename = self.packet.fname
        self.r_opts = self.packet.r_opts
        logging.debug(self.r_opts)
        self.opts = {**self.default_opts, **self.extra_opts, **self.r_opts}
        logging.debug(self.opts)

    def connection_lost(self, exc):
        """
        Cleans up socket and fd after connection has been lost. Logs an error
        if connection interrupted.
        """
        self.conn_reset()
        if exc:
            logging.error(
                'Error on connection lost: {0}.\nTraceback: {1}'.format(
                    exc, exc.__traceback__))
        else:
            logging.info('Connection to {0}:{1} terminated'.format(
                *self.remote_addr))

    def error_received(self, exc):
        """
        Handles cleanup after socket reports an error ie. local or remote
        socket closed and other network errors.
        """

        self.conn_reset()
        self.transport.close()
        logging.error(
            ('Error receiving packet from {0}: {1}. '
             'Transfer of "{2}" aborted.\nTraceback: {3}').format(
                 self.remote_addr, exc, self.filename, exc.__traceback__))

    def send_opening_packet(self, packet):
        """
        Starts the connection timeout timer and sends first datagram.
        """
        self.reply_to_client(packet)
        self.h_timeout = asyncio.get_event_loop().call_later(
            self.opts[b'timeout'], self.conn_timeout)

    def reply_to_client(self, packet):
        """
        Starts the message retry loop, resending packet to self.remote_addr
        every 'ack_timeout'.
        """
        self.transport.sendto(packet, self.remote_addr)
        self.retransmit = asyncio.get_event_loop().call_later(
            self.opts[b'ack_timeout'], self.reply_to_client, packet)

    def handle_err_pkt(self):
        """
        Cleans up connection after sending a courtesy error packet
        to offending client.
        """
        logging.info(('Closing connection to {0} due to error. '
                      '"{1}" Not transmitted.').format(self.remote_addr,
                                                       self.filename))
        self.conn_reset()
        asyncio.get_event_loop().call_soon(self.transport.close)

    def retransmit_reset(self):
        """
        Stops the message retry loop.
        """
        if self.retransmit:
            self.retransmit.cancel()

    def conn_reset(self):
        """
        Stops the message retry loop and the connection timeout timers.
        """
        self.retransmit_reset()
        if self.h_timeout:
            self.h_timeout.cancel()

    def conn_timeout(self):
        """
        Cleans up timers and the connection when called.
        """

        logging.error(
            'Connection to {0} timed out, "{1}" not transfered'.format(
                self.remote_addr, self.filename))
        self.retransmit_reset()
        self.transport.close()

    def conn_timeout_reset(self):
        """
        Restarts the connection timeout timer.
        """

        self.conn_reset()
        self.h_timeout = asyncio.get_event_loop().call_later(
            self.opts[b'timeout'], self.conn_timeout)

    def is_correct_tid(self, addr):
        """
        Checks whether address '(ip, port)' matches that of the
        established remote host.
        May send error to host that submitted incorrect address.
        """
        if self.remote_addr[1] == addr[1]:
            return True
        else:
            logging.info(
                'Unknown transfer id: expected {0}, got {1} instead.'.format(
                    self.remote_addr, addr))
            err_response = self.packet_factory.err_unknown_tid()
            self.transport.sendto(err_response.to_bytes(), addr)
            return False


class WRQProtocol(BaseTFTPProtocol):
    def __init__(self, wrq, file_handler_cls, addr, opts):
        super().__init__(wrq, file_handler_cls, addr, opts)
        logging.info('Initiating WRQProtocol with {0}'.format(
            self.remote_addr))

    def next_datagram(self):
        """
        Builds an acknowledgement of a received data packet.
        """
        return self.packet_factory.create_packet(pkt_type='ACK',
                                                 block_no=self.counter)

    def initialize_transfer(self):
        self.counter = 0
        self.file_handler = self.file_handler_cls(self.filename,
                                                  self.opts[b'blksize'])

    def datagram_received(self, data, addr):
        """
        Check correctness of received datagram, reset timers, increment
        counter, ACKnowledge datagram, save received data to file.
        """
        packet = self.packet_factory.from_bytes(data)

        if (self.is_correct_tid(addr) and packet.is_data() and
                packet.is_correct_sequence((self.counter + 1) % 65536)):
            self.conn_timeout_reset()

            self.counter = (self.counter + 1) % 65536
            reply_packet = self.next_datagram()
            self.reply_to_client(reply_packet.to_bytes())

            self.file_handler.write_chunk(packet.data)

            if packet.size < self.opts[b'blksize']:
                logging.info('Receiving file "{0}" from {1} completed'.format(
                    self.filename, self.remote_addr))
                self.retransmit_reset()
                self.transport.close()
        else:
            logging.debug('Data: {0}; is_data: {1}; counter: {2}'.format(
                data, packet.is_data(), self.counter))


class RRQProtocol(BaseTFTPProtocol):
    def __init__(self, rrq, file_handler_cls, addr, opts):
        super().__init__(rrq, file_handler_cls, addr, opts)
        logging.info('Initiating RRQProtocol with {0}'.format(
            self.remote_addr))

    def next_datagram(self):
        return self.packet_factory.create_packet(pkt_type='DAT',
                                                 block_no=self.counter,
                                                 data=self.file_handler.read_chunk())

    def initialize_transfer(self):
        self.counter = 1
        self.file_handler = self.file_handler_cls(self.filename, self.opts[b'blksize'])

    def datagram_received(self, data, addr):
        """
        Checks correctness of incoming datagrams, reset timers,
        increments message counter, send next chunk of requested file
        to client.
        """
        packet = self.packet_factory.from_bytes(data)
        if (self.is_correct_tid(addr) and packet.is_ack() and
                packet.is_correct_sequence(self.counter)):
            self.conn_timeout_reset()
            self.counter = (self.counter + 1) % 65536
            packet = self.next_datagram()
            self.reply_to_client(packet.to_bytes())
            if self.file_handler.finished:
                self.transport.close()
        else:
            logging.debug('Ack: {0}; is_ack: {1}; counter: {2}'.format(
                data, packet.is_ack(), self.counter))


class BaseTFTPServerProtocol(asyncio.DatagramProtocol):
    def __init__(self, host_interface, loop, extra_opts):
        self.host_interface = host_interface
        self.loop = loop
        self.extra_opts = extra_opts
        self.packet_factory = TFTPPacketFactory()

    def select_protocol(self, request):
        """
        Selects an asyncio.Protocol-compatible protocol to
        feed to an event loop's 'create_datagram_endpoint'
        function.
        """
        raise NotImplementedError

    def select_file_handler(self, first_packet):
        """
        Selects a class that implements the correct interface
        to handle the input/output for a tftp transfer.
        """
        raise NotImplementedError

    def connection_made(self, transport):
        logging.info('Listening...')
        self.transport = transport

    def datagram_received(self, data, addr):
        """
        Opens a read or write connection to remote host by scheduling
        an asyncio.Protocol.
        """
        logging.debug('received: {}'.format(data.decode()))

        first_packet = self.packet_factory.from_bytes(data)
        protocol = self.select_protocol(first_packet)
        file_handler_cls = self.select_file_handler(first_packet)

        connect = self.loop.create_datagram_endpoint(
            lambda: protocol(data, file_handler_cls, addr, self.extra_opts),
            local_addr=(self.host_interface,
                        0, ))

        self.loop.create_task(connect)

    def connection_lost(self, exc):
        logging.info('TFTP server - connection lost')


class TFTPServerProtocol(BaseTFTPServerProtocol):
    def select_protocol(self, packet):
        logging.debug('packet type: {}'.format(packet.pkt_type))
        if packet.is_rrq():
            return RRQProtocol
        elif packet.is_wrq():
            return WRQProtocol
        else:
            raise ProtocolException('Received incompatible request, ignoring.')

    def select_file_handler(self, packet):
        if packet.is_wrq():
            return file_io.FileWriter
        else:
            return file_io.FileReader



class Py3tftpError(Exception):
    pass


class ProtocolException(Py3tftpError):
    pass


class UnacknowledgedOption(Py3tftpError):
    pass


class BadRequest(Py3tftpError):
    pass


class BadPacketType(Py3tftpError):
    pass



from setuptools import setup
from py3tftp import __version__

try:
    import pypandoc
except ImportError:
    pypandoc = None


def readme():
    with open('README.md', 'r') as f:
        readme_md = f.read()
        if pypandoc:
            readme_rst = pypandoc.convert(readme_md, 'rst', format='md')
            return readme_rst
        else:
            return readme_md


setup(
    name='py3tftp',
    version=__version__,
    description='Python 3 asynchronous TFTP server.',
    long_description=readme(),
    url='http://github.com/sirMackk/py3tftp',
    author='Matt O.',
    author_email='matt@mattscodecave.com',
    license='MIT',
    keywords='async asynchronous tftp',
    packages=['py3tftp'],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'py3tftp = py3tftp.__main__:main'
        ]
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Topic :: Utilities',
        'Topic :: Software Development',
        'License :: OSI Approved :: MIT License',
    ],
)
